__version__ = "0.5.1"
__version_info__ = ['0', '5', '1']
__version_vcs_info__ = {
    'branch': 'hotfix-0.5.1',
    'revid': 'b25b422',
    'revno': 393}
__version_user_info__ = "skalkoto@darkstar.admin.grnet.gr"
